/*
 * driver.hh
 *
 *  Created on: Feb 6, 2021
 *      Author: sgupta
 */

#ifndef PROBLEM00_DRIVER_HH_
#define PROBLEM00_DRIVER_HH_

template<typename GV,typename PTree,typename Parameters>
void driver( const GV& gv,const PTree& ptree,Parameters& parameter,
			 std::string pathname, std::string filename,
			 Dune::MPIHelper& helper){

	using Coord = typename GV::Grid::ctype;
	const int dim = GV::dimension;

	/****************************************************/
	// GRID FUNCTION SPACES
	/****************************************************/
	using CON0 = Dune::PDELab::OverlappingConformingDirichletConstraints;
	using VBE0 = Dune::PDELab::ISTL::VectorBackend<> ;

	// FEM SPACE FOR P
	using FEM_P = Dune::PDELab::QkLocalFiniteElementMap<GV,Coord,double,1>;
	FEM_P fem_p(gv);
	using GFS_P = Dune::PDELab::GridFunctionSpace<GV,FEM_P,CON0,VBE0> ;
	GFS_P gfs_p1(gv,fem_p);
	GFS_P gfs_p2(gv,fem_p);

	// FEM SPACE FOR S
#ifdef S_FV
	using FEM_S = Dune::PDELab::QkDGLocalFiniteElementMap<Coord,double,0,dim,Dune::PDELab::QkDGBasisPolynomial::lagrange>;
	FEM_S fem_s;
#else
	using FEM_S = Dune::PDELab::QkLocalFiniteElementMap<GV,Coord,double,1>;
	FEM_S fem_s(gv);
#endif
	using GFS_S = Dune::PDELab::GridFunctionSpace<GV,FEM_S,CON0,VBE0> ;
	GFS_S gfs_s(gv,fem_s);

	// composite GFS
#ifdef S_FV
	using VBE =Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::none>;
#else
	using VBE =Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed>;
#endif
	using OrderingTag = Dune::PDELab::EntityBlockedOrderingTag;
	using GFS = Dune::PDELab::CompositeGridFunctionSpace< VBE,OrderingTag,GFS_P,GFS_P,GFS_S> ;
	GFS gfs(gfs_p1,gfs_p2,gfs_s);
	typedef typename GFS::template ConstraintsContainer<double>::Type C;
	C c;
	c.clear();

	/****************************************************/
	// VECTOR CONTAINERS
	/****************************************************/
	using U = Dune::PDELab::Backend::Vector<GFS,double>;
	U u_old(gfs,0.0);
	U u_new(gfs,0.0);

	/****************************************************/
	// INITIAL CONDITIONS
	/****************************************************/
	using Initial = InitialConditions<GV,Parameters>;
	Initial initial(gv,parameter);
	auto p1_ic_local = [&](const auto& e, const auto& x){return initial.P1(e,x);};
	auto p1_ic = Dune::PDELab::makeGridFunctionFromCallable(gv,p1_ic_local);
	auto p2_ic_local = [&](const auto& e, const auto& x){return initial.P2(e,x);};
	auto p2_ic = Dune::PDELab::makeGridFunctionFromCallable(gv,p2_ic_local);
	auto s_ic_local = [&](const auto& e, const auto& x){return initial.S(e,x);};
	auto s_ic = Dune::PDELab::makeGridFunctionFromCallable(gv,s_ic_local);
	using  InitialConditions = Dune::PDELab::CompositeGridFunction<decltype(p1_ic),decltype(p2_ic),decltype(s_ic)>;
	InitialConditions ic( p1_ic,p2_ic,s_ic );

	// 	Initialize the solution at t=0 (uold) with the given initial values
	Dune::PDELab::interpolate( ic, gfs, u_old );
	u_new = u_old;

	/****************************************************/
	// BOUNDARY CONDITION
	/****************************************************/
	using BC1 = BoundaryCondition1< GV,Parameters>;
	BC1 bc1( gv,parameter);
	using BC2 = BoundaryCondition2< GV,Parameters>;
	BC2 bc2( gv,parameter);

	/****************************************************/
	// LOCAL OPERATORS
	/****************************************************/
	using LOP = LocalOperator<GV,Parameters,BC1,BC2>;
	LOP lop( gv,parameter,bc1,bc2);
	using TOP = TimeOperator<GV,Parameters>;
	TOP tlop(gv, parameter);

	/****************************************************/
	// STATIONARY GRID OPERATOR
	/****************************************************/
	using MBE = Dune::PDELab::ISTL::BCRSMatrixBackend<>;
	MBE mbe(20); // Maximal number of nonzeroes per row

	using GOLOP = Dune::PDELab::GridOperator<GFS,GFS,LOP,MBE,double,double,double,C,C>;
	GOLOP golop(gfs,c,gfs,c,lop,mbe);

	using GOTOP = Dune::PDELab::GridOperator<GFS,GFS,TOP,MBE,double,double,double,C,C>;
	GOTOP gotop(gfs,c,gfs,c,tlop,mbe);

	using IGO = Dune::PDELab::OneStepGridOperator<GOLOP,GOTOP>;
	IGO igo(golop,gotop);

	/****************************************************/
	// SOLVER
	/****************************************************/
	using LS1 = Dune::PDELab::ISTLBackend_SEQ_SuperLU; //works
	LS1 ls1;
	using LS2 = Dune::PDELab::ISTLBackend_BCGS_AMG_SSOR<IGO>;
	LS2 ls2(gfs,1000,1);

	/****************************************************/
	// COUPLED NONLINEAR PROBLEM - INSTATIONARY
	/****************************************************/
	using PDESOLVER1 = Dune::PDELab::NewtonMethod<IGO,LS1>;
	PDESOLVER1 newton1(igo,ls1);
	newton1.setParameters(ptree.sub("newton"));
	using PDESOLVER2 = Dune::PDELab::NewtonMethod<IGO,LS2>;
	PDESOLVER2 newton2(igo,ls2);
	newton2.setParameters(ptree.sub("newton"));

	/****************************************************/
	// TIME STEPPER
	/****************************************************/
    Dune::PDELab::ImplicitEulerParameter<double> method1;
	Dune::PDELab::TimeSteppingParameterInterface<double> *method = & method1;
    using OSM1 = Dune::PDELab::OneStepMethod<double,IGO,PDESOLVER1,U,U>;
    OSM1 osm1(*method,igo,newton1);
    osm1.setVerbosityLevel(1);
    using OSM2 = Dune::PDELab::OneStepMethod<double,IGO,PDESOLVER2,U,U>;
    OSM2 osm2(*method,igo,newton2);
    osm2.setVerbosityLevel(1);

	/****************************************************/
	// VTK
	/****************************************************/
	using VTKWRITER = Dune::VTKWriter<GV>;
	VTKWRITER vtkwriter(gv);
	using VTKSEQUENCEWRITER = Dune::VTKSequenceWriter<GV>;
	VTKSEQUENCEWRITER vtkSequenceWriter(std::make_shared<VTKWRITER>(vtkwriter),filename,pathname,"");

	using SUB_P1 = typename Dune::PDELab::GridFunctionSubSpace< GFS, Dune::TypeTree::HybridTreePath<Dune::index_constant<0>> >;
	SUB_P1 	subgfs_p1(gfs);
	using DGF_P1 = Dune::PDELab::DiscreteGridFunction<SUB_P1,U>;
	DGF_P1 dgf_p1(subgfs_p1,u_new);
	vtkSequenceWriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<DGF_P1> >(dgf_p1,"P1"));

	using SUB_P2 = typename Dune::PDELab::GridFunctionSubSpace< GFS, Dune::TypeTree::HybridTreePath<Dune::index_constant<1>> >;
	SUB_P2 	subgfs_p2(gfs);
	using DGF_P2 = Dune::PDELab::DiscreteGridFunction<SUB_P2,U>;
	DGF_P2 dgf_p2(subgfs_p2,u_new);
	vtkSequenceWriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<DGF_P2> >(dgf_p2,"P2"));

	using SUB_S = typename Dune::PDELab::GridFunctionSubSpace< GFS, Dune::TypeTree::HybridTreePath<Dune::index_constant<2>> >;
	SUB_S 	subgfs_s(gfs);
	using DGF_S = Dune::PDELab::DiscreteGridFunction<SUB_S,U>;
	DGF_S dgf_s(subgfs_s,u_new);
	vtkSequenceWriter.addCellData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<DGF_S> >(dgf_s,"S"));

	vtkSequenceWriter.write(0.0,Dune::VTK::appendedraw);

	/****************************************************/
	// TIME-LOOP
	/****************************************************/
	double time = 0.0;
	double dt = ptree.get("time.dt_initial",(double)0.001);
	double dt_max = ptree.get("time.dt",(double)0.1);
	// total simulation-time
	double t_END  = ptree.get("time.time_end",(double)1000.);
	bool solver_flag = ptree.get("linear_solver.superlu",(bool)true);

	while( time < t_END - 1e-8 ){
		try{
			if(solver_flag)
				osm1.apply( time, dt, u_old, u_new );
			else osm2.apply( time, dt, u_old, u_new );

		}catch ( Dune::Exception &e ) {
			if( dt > 1e-12 ){
				if(helper.rank()==0){
				std::cout << "Catched Error, Dune reported error: " << e << std::endl;
				}
				u_new = u_old;
				dt *= 0.5;
				continue;
			}else{
				if(helper.rank()==0){
				std::cout << "ABORTING, due to DUNE error: " << e << std::endl;
				}
				exit(0);
			}
		}

		if(helper.rank()==0){
		std::cout<<"DONE"<<std::endl;
		std::cout<<"_____________________________________________________"<<std::endl;
		}

		// WRITE OUTPUT
		vtkSequenceWriter.write(time,Dune::VTK::appendedraw);

		/***********************************************/
		// PREPARE FOR NEXT TIME INTEGRATION
		/***********************************************/
		dt = dt_max;
		//1. ASSIGN THE 'NEXT' VALUE TO 'OLD' VARIABLE
		u_old = u_new;
		//2. ADVANCE TIME:
		time += dt;
	}
}

#endif /* PROBLEM00_DRIVER_HH_ */
