/*
 * lop.hh
 *
 *  Created on: Feb 6, 2021
 *      Author: sgupta
 */

#ifndef PROBLEM00_LOP_HH_
#define PROBLEM00_LOP_HH_

template <class GV,class Params,class BC1,class BC2>
class LocalOperator :
  public Dune::PDELab::NumericalJacobianApplyVolume		< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::NumericalJacobianVolume			< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::NumericalJacobianApplySkeleton	< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::NumericalJacobianSkeleton		< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::NumericalJacobianApplyBoundary	< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::NumericalJacobianBoundary		< LocalOperator<GV,Params,BC1,BC2> >,
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<double>
{
private:
    const GV& gv;
	const Params& param;
	const BC1& bc1;
	const BC2& bc2;
	int intorder;
	const static int dim = GV::dimension;

public:
	  // pattern assembly flags
	  enum { doPatternVolume	= true };

	  // residual assembly flags
	  enum { doAlphaVolume  	= true };
	  enum { doAlphaBoundary	= true };

	  // constructor stores parameters
	  LocalOperator( const GV& gv_,
					 const Params& param_,
					 const BC1& bc1_,
					 const BC2& bc2_,
					 int intorder_=4)
	  : gv(gv_),
		param( param_ ),
		bc1( bc1_ ),
		bc2( bc2_ ),
		intorder(intorder_)
	  {}

	  // volume integral depending on test and ansatz functions
	  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
	  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
	  {
		  // subspaces
		  //P1
		  auto lfsv_p1 = lfsv.template child<0>();
		  auto lfsu_p1 = lfsu.template child<0>();

		  auto lfsv_p2 = lfsv.template child<1>();
		  auto lfsu_p2 = lfsu.template child<1>();
		  //S
		  auto lfsv_s = lfsv.template child<2>();
		  auto lfsu_s = lfsu.template child<2>();

		  // define types
		  using RF = typename LFSU::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType::Traits::RangeFieldType;
		  using size_type = typename LFSU::template Child<0>::Type::Traits::SizeType;

		  // local cache
		  using FEM_P = typename LFSU::template Child<0>::Type;
		  using LocalBasis_P = typename FEM_P::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_P = Dune::PDELab::LocalBasisCache<LocalBasis_P>;
		  Cache_P cache_p;

		  using FEM_S = typename LFSU::template Child<2>::Type;
		  using LocalBasis_S = typename FEM_S::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_S = Dune::PDELab::LocalBasisCache<LocalBasis_S>;
		  Cache_S cache_s;

		  // Get geometry
		  auto geo = eg.geometry();

		  // Initialize vectors outside for loop
		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p1(lfsu_p1.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p1(lfsv_p1.size());
		  Dune::FieldVector<RF,dim> grad_p1(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p1(0.0);

		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p2(lfsu_p2.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p2(lfsv_p2.size());
		  Dune::FieldVector<RF,dim> grad_p2(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p2(0.0);

		  // Transformation matrix
		  typename EG::Geometry::JacobianInverseTransposed jac;

		  // loop over quadrature points
		  for (const auto &ip : quadratureRule(geo,intorder))
		  {
			  // evaluate basis functions
			  auto &phi_p1 = cache_p.evaluateFunction(ip.position(), lfsu_p1.finiteElement().localBasis());
			  auto &psi_p1 = cache_p.evaluateFunction(ip.position(), lfsv_p1.finiteElement().localBasis());
			  auto &phi_p2 = cache_p.evaluateFunction(ip.position(), lfsu_p2.finiteElement().localBasis());
			  auto &psi_p2 = cache_p.evaluateFunction(ip.position(), lfsv_p2.finiteElement().localBasis());
			  auto &phi_s = cache_s.evaluateFunction(ip.position(), lfsu_s.finiteElement().localBasis());
			  auto &psi_s = cache_s.evaluateFunction(ip.position(), lfsv_s.finiteElement().localBasis());

			  auto ip_global = geo.global(ip.position());
			  auto ip_local = geo.local(ip_global);

			  // evaluate P1
			  RF P1 = 0.0;
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  P1 += x(lfsu_p1, i) * phi_p1[i];
			  // evaluate P2
			  RF P2 = 0.0;
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  P2 += x(lfsu_p2, i) * phi_p2[i];
			  // evaluate S
			  RF S = 0.0;
			  for (size_type i = 0; i < lfsu_s.size(); i++)
				  S += x(lfsu_s, i) * phi_s[i];

			  // evaluate gradient of basis functions
			  auto &js_u_p1 = cache_p.evaluateJacobian(ip.position(), lfsu_p1.finiteElement().localBasis());
			  auto &js_v_p1 = cache_p.evaluateJacobian(ip.position(), lfsv_p1.finiteElement().localBasis());
			  auto &js_u_p2 = cache_p.evaluateJacobian(ip.position(), lfsu_p2.finiteElement().localBasis());
			  auto &js_v_p2 = cache_p.evaluateJacobian(ip.position(), lfsv_p2.finiteElement().localBasis());

			  // transform gradients of shape functions to real element
			  jac = geo.jacobianInverseTransposed(ip.position());

			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  jac.mv(js_u_p1[i][0], gradphi_p1[i]);
			  for (size_type i = 0; i < lfsv_p1.size(); i++)
				  jac.mv(js_v_p1[i][0], gradpsi_p1[i]);
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  jac.mv(js_u_p2[i][0], gradphi_p2[i]);
			  for (size_type i = 0; i < lfsv_p2.size(); i++)
				  jac.mv(js_v_p2[i][0], gradpsi_p2[i]);

			  //properties
			  auto K1 = param.Permeability1(S);
			  auto K2 = param.Permeability2(S);
			  //source term for PDE
			  auto q_p1 = param.Source_P1(S,P2);
			  auto q_p2 = param.Source_P2(S,P1);
			  auto q_s = param.Source_S(P1,P2);

			  // GRADIENTS
			  // compute KgradP1
			  grad_p1 = 0.0;
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  grad_p1.axpy(x(lfsu_p1, i), gradphi_p1[i]);
			  // compute K * gradient of P
			  K1.mv(grad_p1, Kgrad_p1);

			  // compute KgradP2
			  grad_p2 = 0.0;
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  grad_p2.axpy(x(lfsu_p2, i), gradphi_p2[i]);
			  // compute K * gradient of P
			  K2.mv(grad_p2, Kgrad_p2);

			  // integrate
			  RF factor = ip.weight() * geo.integrationElement(ip.position());
			  // PDE1: \nabla \cdot (-[K1] \nabla P1 ) = q(S,P2);
			  for (size_type i = 0; i < lfsv_p1.size(); i++) {
				  r.accumulate(lfsv_p1, i, ( (Kgrad_p1*gradpsi_p1[i]) + q_p1*psi_p1[i] )*factor );
			  }
			  // PDE2: \nabla \cdot (-[K2] \nabla P2 ) = q(S,P1);
			  for (size_type i = 0; i < lfsv_p2.size(); i++) {
				  r.accumulate(lfsv_p2, i, ( (Kgrad_p2*gradpsi_p2[i]) + q_p2*psi_p2[i] )*factor );
			  }
			  // AE: S-f(P)
			  for (size_type i = 0; i < lfsv_s.size(); i++){
				  r.accumulate(lfsv_s, i, ((S-q_s)*psi_s[i])*factor);
			  }

		  } //End Quadrature Rule
	  }//END:alpha_volume


	  // boundary integral depending on test and ansatz functions
	  template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
	  void alpha_boundary (const IG& ig, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r ) const
	  {
		  // subspaces
		  //P1
		  auto lfsv_p1 = lfsv.template child<0>();
		  auto lfsu_p1 = lfsu.template child<0>();
		  //P2
		  auto lfsv_p2 = lfsv.template child<1>();
		  auto lfsu_p2 = lfsu.template child<1>();
		  //S
		  auto lfsv_s = lfsv.template child<2>();
		  auto lfsu_s = lfsu.template child<2>();

		  // define types
		  using RF = typename LFSU::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType::Traits::RangeFieldType;
		  using size_type = typename LFSU::template Child<0>::Type::Traits::SizeType;

		  // local cache
		  using FEM_P = typename LFSU::template Child<0>::Type;
		  using LocalBasis_P = typename FEM_P::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_P = Dune::PDELab::LocalBasisCache<LocalBasis_P>;
		  Cache_P cache_p;

		  using FEM_S = typename LFSU::template Child<2>::Type;
		  using LocalBasis_S = typename FEM_S::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_S = Dune::PDELab::LocalBasisCache<LocalBasis_S>;
		  Cache_S cache_s;

		  // Get geometries
		  auto intersection = ig.intersection();
		  const auto& cell_inside = ig.inside();
		  auto geo = ig.geometry();
		  auto geo_inside = cell_inside.geometry();
		  auto geo_in_inside = ig.geometryInInside();
		  auto ref_el = referenceElement(geo_in_inside);
		  auto face_center_local = ref_el.position(0,0);

		  // Initialize vectors outside for loop
		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p1(lfsu_p1.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p1(lfsv_p1.size());
		  Dune::FieldVector<RF,dim> grad_p1(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p1(0.0);

		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p2(lfsu_p2.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p2(lfsv_p2.size());
		  Dune::FieldVector<RF,dim> grad_p2(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p2(0.0);

		  // boundary conditions
		  auto bct1 = bc1.type(ig,face_center_local);
		  auto bct2 = bc2.type(ig,face_center_local);

		  // Transformation matrix
		  typename IG::Entity::Geometry::JacobianInverseTransposed jac;

		  // loop over quadrature points
		  // auto intorder = intorderadd+quadrature_factor*order;
		  for (const auto &ip : quadratureRule(geo, intorder))
		  {
			  // exact normal
			  auto n_F_local = ig.unitOuterNormal(ip.position());

			  // position of quadrature point in local coordinates of elements
			  auto iplocal = geo_in_inside.global(ip.position());
			  auto ip_global = geo_inside.global(iplocal);

			  // evaluate basis functions
			  auto &phi_p1 = cache_p.evaluateFunction(iplocal, lfsu_p1.finiteElement().localBasis());
			  auto &psi_p1 = cache_p.evaluateFunction(iplocal, lfsv_p1.finiteElement().localBasis());
			  auto &phi_p2 = cache_p.evaluateFunction(iplocal, lfsu_p2.finiteElement().localBasis());
			  auto &psi_p2 = cache_p.evaluateFunction(iplocal, lfsv_p2.finiteElement().localBasis());
			  auto &phi_s = cache_s.evaluateFunction(iplocal, lfsu_s.finiteElement().localBasis());
			  auto &psi_s = cache_s.evaluateFunction(iplocal, lfsv_s.finiteElement().localBasis());

			  // evaluate P1
			  RF P1 = 0.0;
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  P1 += x(lfsu_p1, i) * phi_p1[i];
			  // evaluate P2
			  RF P2 = 0.0;
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  P2 += x(lfsu_p2, i) * phi_p2[i];
			  // evaluate S
			  RF S = 0.0;
			  for (size_type i = 0; i < lfsu_s.size(); i++)
				  S += x(lfsu_s, i) * phi_s[i];

			  //properties
			  auto K1 = param.Permeability1(S);
			  auto K2 = param.Permeability2(S);

			  // evaluate gradient of basis functions
			  auto &js_u_p1 = cache_p.evaluateJacobian(iplocal, lfsu_p1.finiteElement().localBasis());
			  auto &js_v_p1 = cache_p.evaluateJacobian(iplocal, lfsv_p1.finiteElement().localBasis());
			  auto &js_u_p2 = cache_p.evaluateJacobian(iplocal, lfsu_p2.finiteElement().localBasis());
			  auto &js_v_p2 = cache_p.evaluateJacobian(iplocal, lfsv_p2.finiteElement().localBasis());

			  // transform gradients of shape functions to real element
			  jac = geo_inside.jacobianInverseTransposed(iplocal);
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  jac.mv(js_u_p1[i][0], gradphi_p1[i]);
			  for (size_type i = 0; i < lfsv_p1.size(); i++)
				  jac.mv(js_v_p1[i][0], gradpsi_p1[i]);

			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  jac.mv(js_u_p2[i][0], gradphi_p2[i]);
			  for (size_type i = 0; i < lfsv_p2.size(); i++)
				  jac.mv(js_v_p2[i][0], gradpsi_p2[i]);

			  // GRADIENTS
			  // compute gradient of P1
			  grad_p1 = 0.0;
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  grad_p1.axpy(x(lfsu_p1, i), gradphi_p1[i]);
			  // compute K1 * gradient of P1
			  K1.mv(grad_p1, Kgrad_p1);

			  // compute gradient of P2
			  grad_p2 = 0.0;
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  grad_p2.axpy(x(lfsu_p2, i), gradphi_p2[i]);
			  // compute K2 * gradient of P2
			  K2.mv(grad_p2, Kgrad_p2);

			  /******************************/
			  //INTEGRATION
			  /******************************/
			  // integration factor
			  auto factor = ip.weight() * geo.integrationElement(ip.position());

			  /******************************/
			  // PDE1
			  /******************************/
			  if( bct1==ConvectionDiffusionBoundaryConditions::Outflow ){
				  for (size_type i = 0; i < lfsv_p1.size(); i++){
					  auto o = bc1.outflow_value(intersection,ip.position()) * n_F_local;
					  r.accumulate(lfsv_p1, i, ( (-1.)*(Kgrad_p1*n_F_local)*psi_p1[i] - o*psi_p1[i] )*factor);
				  }
			  }else if(bct1==ConvectionDiffusionBoundaryConditions::Neumann){
				  for (size_type i = 0; i < lfsv_p1.size(); i++){
					  auto j = bc1.neumann_value(intersection,ip.position()) * n_F_local;
					  r.accumulate(lfsv_p1, i, ( -j*psi_p1[i] )*factor);
				  }
			  }else{
				  for (size_type i = 0; i < lfsv_p1.size(); i++){
					  auto P1_n = bc1.dirichlet_value(intersection,ip.position());
					  r.accumulate(lfsv_p1, i, ( (P1-P1_n)*psi_p1[i] )*factor);
				  }
			  }

			  /******************************/
			  // PDE2
			  /******************************/
			  if( bct2==ConvectionDiffusionBoundaryConditions::Outflow ){
				  for (size_type i = 0; i < lfsv_p2.size(); i++){
					  auto o = bc2.outflow_value(intersection,ip.position()) * n_F_local;
					  r.accumulate(lfsv_p2, i, ( (-1.)*(Kgrad_p2*n_F_local)*psi_p2[i] - o*psi_p2[i] )*factor);
				  }
			  }else if(bct2==ConvectionDiffusionBoundaryConditions::Neumann){
				  for (size_type i = 0; i < lfsv_p2.size(); i++){
					  auto j = bc2.neumann_value(intersection,ip.position()) * n_F_local;
					  r.accumulate(lfsv_p2, i, ( -j*psi_p2[i] )*factor);
				  }
			  }else{
				  for (size_type i = 0; i < lfsv_p2.size(); i++){
					  auto P2_n = bc2.dirichlet_value(intersection,ip.position());
					  r.accumulate(lfsv_p2, i, ( (P2-P2_n)*psi_p2[i] )*factor);
				  }
			  }

		  } //End Quadrature Rule

	  }//END:alpha_boundary

};

#endif /* PROBLEM00_LOP_HH_ */
