/*
 * parameters.hh
 *
 *  Created on: Feb 6, 2021
 *      Author: sgupta
 */

#ifndef PROBLEM00_PARAMETERS_HH_
#define PROBLEM00_PARAMETERS_HH_

template<typename GV,typename PTree>
class DefaultParameters
{
private:
	  const GV& gv;
	  const PTree& ptree;
	  const static int dim = GV::dimension;
	  double eps = 1.e-6;

	  bool VGflag;
	  double Pmax;
	  double K0;
	  double lambda;
	  double alpha;
	  double m;
	  double q1;
	  double mu1;
	  double P1_L;
	  double P1_R;
	  double P1_t0;
	  double q2;
	  double mu2;
	  double P2_t0;
	  double P2_L;
	  double P2_R;
	  double L;

public:
	  //! construct from grid view
	  DefaultParameters (const GV& gv_, const PTree& ptree_)
	: gv( gv_ ) ,
	  ptree(ptree_)
	{
		L = ptree.get("domain.length.x",(double)1.);
		K0 = ptree.get("parameters.K" ,(double)1.e-12);
		VGflag = ptree.get("parameters.van_genuchten.flag",(bool)false);
		lambda = ptree.get("parameters.van_genuchten.lambda" ,(double)2.0);
		alpha = ptree.get("parameters.van_genuchten.alpha" ,(double)0.00015);
		m = ptree.get("parameters.van_genuchten.m" ,(double)0.6);
		Pmax = ptree.get("parameters.Pmax" ,(double)1.e8);
		q1 = ptree.get("parameters.source1" ,(double)0.0);
		q2 = ptree.get("parameters.source2" ,(double)0.0);
		mu1 =ptree.get("parameters.mu1" ,(double)1.e-3);
		mu2 =ptree.get("parameters.mu2" ,(double)1.e-5);
		P1_t0 = ptree.get("initial.P1" ,(double)1.e5);
		P2_t0 = ptree.get("initial.P2" ,(double)0.8e5);
		P1_L = ptree.get("boundary.P1.left" ,(double)2.e5);
		P1_R = ptree.get("boundary.P1.right",(double)1.e5);
		P2_L = ptree.get("boundary.P2.left" ,(double)2.e5);
		P2_R = ptree.get("boundary.P2.right",(double)1.e5);
	}

	bool isLeft(Dune::FieldVector<double,dim> globalpos)const{
  		if(globalpos[0]<0.+eps) return true;
  		else return false;
	}

  	bool isRight(Dune::FieldVector<double,dim> globalpos) const {
  		if(globalpos[0]>L-eps) return true;
  		else return false;
  	}

	Dune::FieldMatrix<double,dim,dim>
	Permeability1(double S) const {
		double kr = S;
		if(VGflag){
			kr = std::pow(S, (2.0/lambda + 3.0) );
		}
		Dune::FieldMatrix<double,dim,dim> K(0.);
		for(int i=0; i<dim; i++){
			K[i][i]= K0*kr/mu1;
		}
		return K;
	}

	Dune::FieldMatrix<double,dim,dim>
	Permeability2(double S) const {
		double kr = (1.-S);
		if(VGflag){
			kr = std::pow(1.-S, 2.) * ( 1.0 - std::pow(S, (2./lambda + 1.) ) );
		}
		Dune::FieldMatrix<double,dim,dim> K(0.);
		for(int i=0; i<dim; i++){
			K[i][i]= K0*kr/mu2;
		}
		return K;
	}

	double Source_S(double P1, double P2)const{
		auto Pc=std::max(P2-P1,0.);
		if(VGflag){
			double S1 = std::pow( std::pow( alpha*Pc , 1./(1.-m) ) + 1. , -m );
			double S2 = 1.-((P2-P1)/Pmax);
			return std::min(S1,S2);
		}else{
			double S = 1.-(Pc/Pmax);
			return S;
		}
	}

	double Source_P1(double S, double P2)const{
		return q1;
	}

	double Source_P2(double S, double P1)const{
		return q2;
	}

	double InitialP1()const{
		return P1_t0;
	}

	double InitialP2()const{
		return P2_t0;
	}

	double InitialS()const{
		return Source_S(P1_t0,P2_t0);
	}

	double P1_Left()const{
		return P1_L;
	}

	double P1_Right()const{
		return P1_R;
	}

	double P2_Left()const{
		return P2_L;
	}

	double P2_Right()const{
		return P2_R;
	}

};


template<typename GV,typename Parameters>
class InitialConditions
{
private:
	const GV& gv;
	const Parameters& param;
	const static int dim = GV::dimension;

public:
	//! construct from grid view
	InitialConditions ( const GV& gv_ ,
						const Parameters& param_)
	: gv( gv_ ) ,
	  param(param_)
	{}

	template<typename E, typename X>
	double P1 (const E& e, const X& x) const {
		return param.InitialP1();
	}

	template<typename E, typename X>
	double P2 (const E& e, const X& x) const {
		return param.InitialP2();
	}

	template<typename E, typename X>
	double S (const E& e, const X& x) const {
		return param.InitialS();
	}

};



struct ConvectionDiffusionBoundaryConditions
{
  enum Type { Dirichlet=1, Neumann=-1, Outflow=-2 };
};

template<typename GV,typename Parameters>
class BoundaryCondition1
{
	typedef ConvectionDiffusionBoundaryConditions::Type BCType;

private:
	const GV& gv;
	const Parameters& param;
	const static int dim = GV::dimension;

public:
	//! construct from grid view
	BoundaryCondition1( const GV& gv_ ,
						const Parameters& param_)
	: gv( gv_ ),param(param_)
	{}

	//! boundary condition type function (true = Dirichlet)
	template<typename I, typename X>
	BCType type (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		if( param.isLeft(xglobal) or param.isRight(xglobal) )
			return ConvectionDiffusionBoundaryConditions::Dirichlet;
		else return ConvectionDiffusionBoundaryConditions::Neumann;
	}

	//! Dirichlet value
	//! Specifies P|_D
	template<typename I, typename X>
	double dirichlet_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		if( param.isLeft(xglobal) )
			return param.P1_Left();
		else return param.P1_Right();
	}

	//! Neumann boundary flux
	template<typename I, typename X>
	Dune::FieldVector<double,dim>
	neumann_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		Dune::FieldVector<double,dim> jF(0.0);
		return jF;
	}

	//! Outflow boundary flux
	template<typename I, typename X>
	Dune::FieldVector<double,dim>
	outflow_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		Dune::FieldVector<double,dim> oF(0.0);
		return oF;
	}

};

template<typename GV,typename Parameters>
class BoundaryCondition2
{
	typedef ConvectionDiffusionBoundaryConditions::Type BCType;

private:
	const GV& gv;
	const Parameters& param;
	const static int dim = GV::dimension;

public:
	//! construct from grid view
	BoundaryCondition2( const GV& gv_ ,
						const Parameters& param_)
	: gv( gv_ ),param(param_)
	{}

	//! boundary condition type function (true = Dirichlet)
	template<typename I, typename X>
	BCType type (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		if( param.isLeft(xglobal) or param.isRight(xglobal) )
			return ConvectionDiffusionBoundaryConditions::Dirichlet;
		else return ConvectionDiffusionBoundaryConditions::Neumann;
	}

	//! Dirichlet value
	//! Specifies P|_D
	template<typename I, typename X>
	double dirichlet_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		if( param.isLeft(xglobal) )
			return param.P2_Left();
		else return param.P2_Right();
	}

	//! Neumann boundary flux
	template<typename I, typename X>
	Dune::FieldVector<double,dim>
	neumann_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		Dune::FieldVector<double,dim> jF(0.0);
		return jF;
	}

	//! Outflow boundary flux
	template<typename I, typename X>
	Dune::FieldVector<double,dim>
	outflow_value (const I& i, const X& x) const {
		auto xglobal = i.geometry().global(x);
		Dune::FieldVector<double,dim> oF(0.0);
		return oF;
	}

};



#endif /* PROBLEM00_PARAMETERS_HH_ */
