/*
 * top.hh
 *
 *  Created on: Feb 6, 2021
 *      Author: sgupta
 */

#ifndef PROBLEM00_TOP_HH_
#define PROBLEM00_TOP_HH_

template < class GV,class Params >
class TimeOperator:
    public Dune::PDELab::NumericalJacobianApplyVolume< TimeOperator<GV,Params> >,
    public Dune::PDELab::NumericalJacobianVolume	 < TimeOperator<GV,Params> >,
	public Dune::PDELab::FullVolumePattern,
    public Dune::PDELab::LocalOperatorDefaultFlags,
    public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<double>
{
private:
	const GV& gv;
	const Params& param;
	unsigned int intorder ;
	const static int dim = GV::dimension;

public:
	// pattern assembly flags
	enum { doPatternVolume = true };

	// residual assembly flags
	enum { doAlphaVolume = true };

	// constructor remembers parameters
	TimeOperator( const GV& gv_, const Params& param_, unsigned int	intorder_ = 4)
	:	gv(gv_), param(param_), intorder( intorder_ )
	{}

	// volume integral depending on test and ansatz functions
	template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
	void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
	{
		  // subspaces
		  //P1
		  auto lfsv_p1 = lfsv.template child<0>();
		  auto lfsu_p1 = lfsu.template child<0>();

		  auto lfsv_p2 = lfsv.template child<1>();
		  auto lfsu_p2 = lfsu.template child<1>();
		  //S
		  auto lfsv_s = lfsv.template child<2>();
		  auto lfsu_s = lfsu.template child<2>();

		  // define types
		  using RF = typename LFSU::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType::Traits::RangeFieldType;
		  using size_type = typename LFSU::template Child<0>::Type::Traits::SizeType;

		  // local cache
		  using FEM_P = typename LFSU::template Child<0>::Type;
		  using LocalBasis_P = typename FEM_P::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_P = Dune::PDELab::LocalBasisCache<LocalBasis_P>;
		  Cache_P cache_p;

		  using FEM_S = typename LFSU::template Child<2>::Type;
		  using LocalBasis_S = typename FEM_S::Traits::FiniteElementType::Traits::LocalBasisType;
		  using Cache_S = Dune::PDELab::LocalBasisCache<LocalBasis_S>;
		  Cache_S cache_s;

		  // Get geometry
		  auto geo = eg.geometry();

		  // Initialize vectors outside for loop
		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p1(lfsu_p1.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p1(lfsv_p1.size());
		  Dune::FieldVector<RF,dim> grad_p1(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p1(0.0);

		  std::vector<Dune::FieldVector<RF,dim>> gradphi_p2(lfsu_p2.size());
		  std::vector<Dune::FieldVector<RF,dim>> gradpsi_p2(lfsv_p2.size());
		  Dune::FieldVector<RF,dim> grad_p2(0.0);
		  Dune::FieldVector<RF,dim> Kgrad_p2(0.0);

		  // Transformation matrix
		  typename EG::Geometry::JacobianInverseTransposed jac;

		  // loop over quadrature points
		  for (const auto &ip : quadratureRule(geo,intorder))
		  {
			  // evaluate basis functions
			  auto &phi_p1 = cache_p.evaluateFunction(ip.position(), lfsu_p1.finiteElement().localBasis());
			  auto &psi_p1 = cache_p.evaluateFunction(ip.position(), lfsv_p1.finiteElement().localBasis());
			  auto &phi_p2 = cache_p.evaluateFunction(ip.position(), lfsu_p2.finiteElement().localBasis());
			  auto &psi_p2 = cache_p.evaluateFunction(ip.position(), lfsv_p2.finiteElement().localBasis());
			  auto &phi_s = cache_s.evaluateFunction(ip.position(), lfsu_s.finiteElement().localBasis());
			  auto &psi_s = cache_s.evaluateFunction(ip.position(), lfsv_s.finiteElement().localBasis());

			  auto ip_global = geo.global(ip.position());
			  auto ip_local = geo.local(ip_global);

			  // evaluate P1
			  RF P1 = 0.0;
			  for (size_type i = 0; i < lfsu_p1.size(); i++)
				  P1 += x(lfsu_p1, i) * phi_p1[i];
			  // evaluate P2
			  RF P2 = 0.0;
			  for (size_type i = 0; i < lfsu_p2.size(); i++)
				  P2 += x(lfsu_p2, i) * phi_p2[i];
			  // evaluate S
			  RF S = 0.0;
			  for (size_type i = 0; i < lfsu_s.size(); i++)
				  S += x(lfsu_s, i) * phi_s[i];

	    	// integration
	    	RF factor = ip.weight() * geo.integrationElement(ip.position());
	    	// PDE1
	    	for (size_type i = 0; i < lfsv_p1.size(); i++) {
	    		r.accumulate(lfsv_p1, i, S * psi_p1[i] * factor);
	    	}
	    	// PDE2
	    	for (size_type i = 0; i < lfsv_p2.size(); i++) {
	    		r.accumulate(lfsv_p2, i, (1-S) * psi_p2[i] * factor);
	    	}

		}

	}// END:alpha_volume

};
#endif /* PROBLEM00_TOP_HH_ */
